"""
    ...
"""
from qt import *
import operator
import config
from util import find, exists

class MyPoint:
    def __init__(self, x, y):
        self.x = x
        self.y = y
        
class FlashingLabel(QLabel):
    def __init__(self, text, parent):
        QLabel.__init__(self, text, parent)
        low = 0.5
        high = 0.5
        pal = self.palette().active()
        # BEGIN works with plastik but not keramik
        #red = pal.text().red()*low + pal.highlightedText().red()*high
        #green = pal.text().green()*low + pal.highlightedText().green()*high
        #blue = pal.text().blue()*low + pal.highlightedText().blue()*high
        # END
        self.highlightColor = QColor(255 - pal.text().red(), 255- pal.text().green()
                                     , 255 - pal.text().blue())
    
    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(150)
    def stopFlashing(self, resetColor):
        self.killTimers()
        if resetColor:
            self.flashingState = FlashingStateNo()
            self.setPaletteForegroundColor(self.palette().active().text())
        else:
            self.flashingState = FlashingStateYesStill()
            self.setPaletteForegroundColor(self.highlightColor)
            
    def timerEvent(self, ev):
        mod = self.flashingState.remainingFlashesBeforeStop % 6
        if mod != 0:
            self.setPaletteForegroundColor(self.palette().active().text())
        else:
            self.setPaletteForegroundColor(self.highlightColor)

        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)


class FlashingTextEdit(QTextEdit):
    def __init__(self, parent):
        QTextEdit.__init__(self, parent)
        self.originalBackgroundColor = self.paletteBackgroundColor()
        self.originalForegroundColor = self.paletteForegroundColor()
    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(
            howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(150)
    def stopFlashing(self, resetColor):
        self.killTimers()
        if resetColor:
            self.flashingState = FlashingStateNo()
            self.setPaletteBackgroundColor(self.originalBackgroundColor)
            self.setPaletteForegroundColor(self.originalForegroundColor)
        else:
            self.flashingState = FlashingStateYesStill()
            self.setPaletteBackgroundColor(self.palette().active().highlight())
            self.setPaletteForegroundColor(self.palette().active().highlightedText())
            
    def timerEvent(self, ev):
        mod = self.flashingState.remainingFlashesBeforeStop % 2
        if mod == 0:
            self.setPaletteBackgroundColor(self.palette().active().light())
        else:
            self.setPaletteBackgroundColor(self.palette().active().mid())

        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)
    
    

class FlashingButton(QPushButton):
    def __init__(self, icon, text, parent):
        QPushButton.__init__(self, icon, text, parent)
        self.originalBackgroundColor = self.paletteBackgroundColor()
        self.originalForegroundColor = self.paletteForegroundColor()
    
    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(
            howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(150)
    def stopFlashing(self, resetColor = True):
        self.killTimers()
        if resetColor:
            self.flashingState = FlashingStateNo()
            self.setPaletteBackgroundColor(self.originalBackgroundColor)
            self.setPaletteForegroundColor(self.originalForegroundColor)
        else:
            self.flashingState = FlashingStateYesStill()
            self.setPaletteBackgroundColor(self.palette().active().highlight())
            self.setPaletteForegroundColor(self.palette().active().highlightedText())
            
    def timerEvent(self, ev):
        mod = self.flashingState.remainingFlashesBeforeStop % 2
        if mod == 0:
            self.setPaletteBackgroundColor(self.palette().active().mid()) # new plastik seems to prefer dark
        else:
            self.setPaletteBackgroundColor(self.palette().active().light())

        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)


class ListArea(QWidget):
    def __init__(self, parentWidget, parentList):
        QWidget.__init__(self, parentWidget)
        self.parentList = parentList
        self.setMouseTracking(True)
        self.setBackgroundMode(Qt.NoBackground)

    def wheelEvent(self, ev):
        movedUp = ev.delta() > 0
        self.parentList.centeringState = None
        if movedUp:
            self.parentList.scrollUpLineWithoutUpdating()
        else:
            self.parentList.scrollDownLineWithoutUpdating()
        self.parentList.update()
        
    def paintEvent(self, ev):
        #print "ListArea.paintEvent. pos = ", point_to_string(self.parentList.lastMousePos)
        self.paint(self.parentList.lastMousePos)

    def paint(self, mousePos):
        #print "ListArea.paint called with mousePos = ", point_to_string(mousePos)
        pix = QPixmap(self.size())
        painter = QPainter(pix) # begin painting on the pixmap
        pix.fill(self, QPoint(0,0))
        currentY = 0
        i = self.parentList.scrollingOffset
        ht = self.height()
        while i < len(self.parentList.items) and currentY < ht:
            item = self.parentList.items[i]
            item.draw(painter, currentY, self.style(),
                  self.colorGroup(),self.palette(),
                  mousePos)
            currentY += item.height
            i += 1
        painter.end() # stop painting on pix
        painter.begin(self) # begin painting on the listarea
        painter.drawPixmap(0,0,pix)
        
    def resizeEvent(self, ev):
        for item in self.parentList.items:
            item.widthChanged(ev.size().width())
        if not (self.parentList.centeringState is None):
            self.parentList.centerItem(self.parentList.centeringState.item,
                self.parentList.centeringState.yoffset)
        self.paint(self.parentList.lastMousePos)
        self.parentList.updateScrollButtonArrows()
        
    def mouseMoveEvent(self, ev):
        #print "mouseMoveEvent. setting new pos = ", point_to_string(ev.pos())
        self.parentList.lastMousePos = MyPoint(ev.pos().x(), ev.pos().y())
        self.paint(self.parentList.lastMousePos)
        
    def leaveEvent(self, ev):
        #print "leaveEvent. setting new pos = ", None
        self.parentList.lastMousePos = None
        self.paint(None)
        
    def mousePressEvent(self, ev):
        curY = 0
        self.lastMousePos = MyPoint(ev.pos().x(), ev.pos().y())
        for item in self.parentList.items[self.parentList.scrollingOffset:]:
            item.ifYouHaveBeenClickedEmitAppropriateSignal(mousePos = self.lastMousePos,
                startY = curY,
                signalEmitter = self.parentList)
            curY += item.height

            
class TextSeparator:
    def __init__(self, text, width):
        self.text = text
        self.widthChanged(width)
    def ifYouHaveBeenClickedEmitAppropriateSignal(self, mousePos, startY, signalEmitter):
        pass
    def widthChanged(self, wt):
        self.width = wt
        # BEGIN calculate height and textHeight
        f = QFont()
        qfontm = QFontMetrics(f)
        boundingRectangle = qfontm.boundingRect(0,0,wt,1000,Qt.BreakAnywhere, self.text)
        self.height = boundingRectangle.height() + config.gSpaceAboveTextSeparator + config.gSpaceBelowTextSeparator
        self.textHeight = boundingRectangle.height()
        # END
        
    def draw(self, painter, y, style, colorGroup, palette, mouse):
        painter.setPen(palette.active().link())
        painter.drawText(0, y + config.gSpaceAboveTextSeparator, self.width, self.height, 
                 Qt.BreakAnywhere, self.text)


class text_item_button_info:
    def __init__(self, pixmaps, meaning, pushbutton_info):
        self.pixmaps = pixmaps
        self.meaning = meaning
        self.pushbutton_info = pushbutton_info
        
class TextItem:
    def __init__(self, text, icon, buttons_info, isSelectable, width, pressed, data):
        self.text  = text
        assert isinstance(icon, QPixmap)
        self.icon = icon
        self.buttons_info = buttons_info
        self.isSelectable = isSelectable
        self.pressed = pressed
        self.widthChanged(width)
        self.data = data
        self.button_pushed = {} # map int -> bool
        for i in range(0, len(self.buttons_info)):
            inf = self.buttons_info[i]
            if cmp(inf.pushbutton_info, "not a pushbutton") == 0:
                pass
            elif cmp(inf.pushbutton_info , "pressed") == 0:
                self.button_pushed[i] = True
            else:
                self.button_pushed[i] = False
                
    def widthChanged(self, w):
        self.width = w
        # BEGIN calculate self.total_buttons_wt
        self.total_buttons_wt = config.g_text_item_button_width * len(self.buttons_info)
        # END
        
        # BEGIN calculate availableWtForText
        availableWtForText = w - config.gItemIconWt - self.total_buttons_wt
        if availableWtForText < 1:
            availableWtForText = 1
        # END
        
        # BEGIN calculate self.height and self.YOffsetForDrawingText
        f = QFont()
        qfontm = QFontMetrics(f)
        boundingRectangle = qfontm.boundingRect(0,0, availableWtForText,
            1000, Qt.BreakAnywhere, self.text)
        if config.gMinItemHt > boundingRectangle.height():
            self.height = config.gMinItemHt
            self.YOffsetForDrawingText = (config.gMinItemHt - boundingRectangle.height())/2
        else:
            if boundingRectangle.height() > 26:
                self.height= boundingRectangle.height() + 8
                self.YOffsetForDrawingText = 4
            else:
                self.height= boundingRectangle.height()
                self.YOffsetForDrawingText = 0
        # END
    def ifYouHaveBeenClickedEmitAppropriateSignal(self, mousePos, startY, signalEmitter):
        if self.isSelectable and self.thePointIsInsideMyText(point = mousePos, myStartY = startY):
            signalEmitter.emit(PYSIGNAL("item_clicked"), (signalEmitter, self))
        res = self.button_icon_the_pointer_is_inside(point = mousePos, myStartY = startY)
        if res != None:
            (i,meaning) = res
            if cmp(self.buttons_info[i].pushbutton_info, "not a pushbutton") != 0:
                if self.button_pushed.get(i, False) == False:
                    self.button_pushed[i] = True
                    signalEmitter.emit(PYSIGNAL("item_button_pressed"), (signalEmitter, self, meaning))
                else:
                    self.button_pushed[i] = False
                    signalEmitter.emit(PYSIGNAL("item_button_released"), (signalEmitter, self, meaning))
            else:
                signalEmitter.emit(PYSIGNAL("item_button_clicked"), (signalEmitter, self, meaning))

    def thePointIsInsideMyText(self, point, myStartY):
        textRightBound = self.width - self.total_buttons_wt
        return point.x < textRightBound and point.y >= myStartY and point.y < myStartY+ self.height
    #    def thePointIsInsideSomeButtonIcon(self, point, myStartY):
    #        return exists(range(0, len(self.buttons_info))
    #                      ,lambda i: i == self.button_icon_the_pointer_is_inside(point, myStartY))
    def button_icon_the_pointer_is_inside(self, point, myStartY):
        textRightBound = self.width - self.total_buttons_wt
        for i in range(0, len(self.buttons_info)):
            left_bound_icon_i = textRightBound + i * config.g_text_item_button_width
            right_bound_icon_i = left_bound_icon_i + config.g_text_item_button_width
            if ( point.x >= left_bound_icon_i and point.x < right_bound_icon_i
                 and point.y >= myStartY and point.y < myStartY+self.height):
                but_info = self.buttons_info[i]
                return (i, but_info.meaning)
        return None
    def draw(self, painter, y, style, colorGroup, palette, mouse):
        textRightBound = self.width - self.total_buttons_wt
        # BEGIN draw the mouseover highlight around text and buttons
        if self.pressed:
            r = QRect(0, y, textRightBound, self.height)
            style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                        QStyle.Style_Enabled | QStyle.Style_Down)
        if (not self.pressed and self.isSelectable and not mouse is None
             and self.thePointIsInsideMyText(point = mouse, myStartY = y)):
            r = QRect(0, y, textRightBound, self.height)
            style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                        QStyle.Style_Enabled | QStyle.Style_Up)
            painter.setPen(Qt.SolidLine)
        if mouse != None:
            result = self.button_icon_the_pointer_is_inside(point = mouse, myStartY = y)
            if  result != None:
                (ic, meaning) = result
                r2 = QRect(0, y, textRightBound, self.height)
                painter.fillRect(r2, QBrush(palette.active().light()));
                r = QRect(textRightBound + ic * config.g_text_item_button_width
                          , y, config.g_text_item_button_width, self.height)
                style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                            QStyle.Style_Enabled | QStyle.Style_Up)
                painter.setPen(Qt.SolidLine)
        # END
        # BEGIN draw the pressed button
        for i in range(0, len(self.buttons_info)):
            if self.button_pushed.get(i, False):
                r = QRect(textRightBound + i * config.g_text_item_button_width
                          , y, config.g_text_item_button_width, self.height)
                style.drawPrimitive(QStyle.PE_ButtonCommand, painter, r, colorGroup,
                                    QStyle.Style_Enabled | QStyle.Style_Down)

        # END
        
        # BEGIN draw text
        if self.isSelectable:
            painter.setPen(palette.active().text())
        else:
            bg = 0.4
            text = 0.6
            red = palette.active().text().red()*text + palette.active().background().red()*bg
            green = palette.active().text().green()*text + palette.active().background().green()*bg
            blue = palette.active().text().blue()*text + palette.active().background().blue()*bg
            painter.setPen(QColor(red, green, blue))
        painter.drawText(config.gItemIconWt, y+self.YOffsetForDrawingText,
                         self.width - config.gItemIconWt - self.total_buttons_wt,
                         self.height, Qt.BreakAnywhere, self.text)
        # END

        # BEGIN draw button icons
        for i in range(0,len(self.buttons_info)):
            # BEGIN find which pixmap to use
            inf = self.buttons_info[i]
            if mouse != None:
                result = self.button_icon_the_pointer_is_inside(point = mouse, myStartY = y)
                if  result != None:
                    (index, meaning) = result
                    if index == i:
                        pix = inf.pixmaps[1]
                    else:
                        pix = inf.pixmaps[0]
                else:
                    pix = inf.pixmaps[0]
            else:
                pix = inf.pixmaps[0]
            # END
            assert pix.width() < config.g_text_item_button_width
            assert self.height > pix.height()
            gap_x =  config.g_text_item_button_width - pix.width()
            gap_y =  self.height - pix.height()
            painter.drawPixmap(textRightBound + i
                               * config.g_text_item_button_width + gap_x/2
                               , y + gap_y/2, pix)
        # END
            
        # BEGIN draw leftmost icon
        if self.icon.height() > self.height:
            iconYOffset = 0
        else:
            iconYOffset = (self.height - self.icon.height())/2
        if self.icon.width() > config.gItemIconWt:
            iconXOffset = 0
        else:
            iconXOffset = (config.gItemIconWt - self.icon.width())/2
        painter.drawPixmap(QPoint(iconXOffset, y + iconYOffset), self.icon)
        # END

class CenteringCommand:
    def __init__(self, box, itemText, yoffset):
        self.itemText = itemText
        self.box = box
        self.yoffset = yoffset
    
class CenteringState:
    def __init__(self, item, yoffset):
        self.item = item
        self.yoffset = yoffset


class FlashingStateYesAnimating:
    def __init__(self, remainingFlashesBeforeStop, resetOriginalColorsWhenFinished):
        self.resetOriginalColorsWhenFinished = resetOriginalColorsWhenFinished
        self.remainingFlashesBeforeStop = remainingFlashesBeforeStop
class FlashingStateYesStill:
    def __init__(self):
        pass
class FlashingStateNo:
    def __init__(self):
        pass
        
class MyListBox(QHBox):
    def enable_all_items(self, selectable):
        for i in self.items:
            if isinstance(i, TextItem):
                i.isSelectable = selectable
    def set_pushbuttons_of_this_item_if_it_exists(self,  pressed, item_data):
        it = find(self.items
                  , lambda i: isinstance(i, TextItem) and i.data == item_data)
        if it != None:
            for b in range(0, len(it.buttons_info)):
                if cmp(it.buttons_info[b].pushbutton_info, "not a pushbutton") == 0:
                    pass
                else:
                    it.button_pushed[b] = pressed
    def enable_these_items(self, items):
        for i in self.items:
            if isinstance(i, TextItem):
                if i.data in items:
                    i.isSelectable = True
                else:
                    i.isSelectable = False

    def isFlashing(self):
        return not isinstance(self.flashingState, FlashingStateNo)

    def startFlashing(self, howManyFlashes, resetOriginalColorsWhenFinished):
        self.stopFlashing(resetColor = True)
        self.flashingState = FlashingStateYesAnimating(howManyFlashes, resetOriginalColorsWhenFinished)
        self.startTimer(100)

    def stopFlashing(self, resetColor):
        self.killTimers()
        if resetColor:
            self.flashingState = FlashingStateNo()
            self.vbox.setPaletteBackgroundColor(self.palette().active().background())
        else:
            self.flashingState = FlashingStateYesStill()
            self.vbox.setPaletteBackgroundColor(self.palette().active().highlight())
        
    def __init__(self, frameVisible, parent):
        QHBox.__init__(self, parent)
        self.items = []
        self.centeringState = None
        self.frameVisible = frameVisible

        # BEGIN create ListArea
        self.vbox = QVBox(self)
        self.vbox.show()
        self.vbox.setMargin(config.gListBoxMarginOfFlashingVBox)
        if self.frameVisible:
            self.vbox.setFrameStyle(QFrame.StyledPanel)
        vbox2 = QVBox(self.vbox)
        vbox2.show()
        vbox2.setMargin(config.gListBoxMarginOfNonFlashingVBox)
        self.listArea = ListArea(parentList = self, parentWidget = vbox2)
        self.listArea.show()
        self.listArea.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Ignored)
        # END

        wt = 40
        ht = 40
        
        self.upAndDownButtonBox = QVBox(self)
        self.upAndDownButtonBox.show()
        self.upAndDownButtonBox.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.upAndDownButtonBox.setMaximumWidth(wt)
        
        # BEGIN up button box
        self.buttonBoxUp = QVBox(self.upAndDownButtonBox)
        self.buttonBoxUp.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.buttonBoxUp.show()

        self.gotoTopBut = QPushButton(self.buttonBoxUp)
        self.gotoTopBut.show()
        self.gotoTopBut.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.gotoTopBut.setMaximumHeight(ht)
        QObject.connect(self.gotoTopBut, SIGNAL("clicked()"), self.onGotoTopClicked)

        self.buttonUp = QPushButton(self.buttonBoxUp)
        self.buttonUp.show()
        self.buttonUp.setMinimumWidth(wt)
        self.buttonUp.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        QObject.connect(self.buttonUp, SIGNAL("clicked()"), self.onPageUpClicked)
        # END

        # BEGIN down button box
        self.buttonBoxDown = QVBox(self.upAndDownButtonBox)
        self.buttonBoxDown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.buttonBoxDown.show()

        self.buttonDown = QPushButton(self.buttonBoxDown)
        self.buttonDown.show()
        self.buttonDown.setMinimumWidth(wt)
        self.buttonDown.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        QObject.connect(self.buttonDown, SIGNAL("clicked()"), self.onPageDownClicked)
        
        self.gotoBottomBut = QPushButton(self.buttonBoxDown)
        self.gotoBottomBut.show()
        self.gotoBottomBut.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.gotoBottomBut.setMaximumHeight(ht)
        QObject.connect(self.gotoBottomBut, SIGNAL("clicked()"), self.onGotoBottomClicked)
        # END
        
        self.setSpacing(5)
        self.scrollingOffset = 0
        self.lastMousePos = None

    def timerEvent(self, ev):
        #print "enter timerEvent. mousepos = ", point_to_string(self.lastMousePos)
        mod = self.flashingState.remainingFlashesBeforeStop % 4
        if mod == 0 or mod == 2:
            self.vbox.setPaletteBackgroundColor(self.palette().active().mid())
        elif mod == 1:
            self.vbox.setPaletteBackgroundColor(self.palette().active().light())
        else:
            self.vbox.setPaletteBackgroundColor(self.palette().active().dark())
        self.flashingState.remainingFlashesBeforeStop -= 1
        if self.flashingState.remainingFlashesBeforeStop <= 0:
            self.stopFlashing(self.flashingState.resetOriginalColorsWhenFinished)
        #print "exit timerEvent. mousepos = ", point_to_string(self.lastMousePos)
        
    def clear(self):
        self.items = []
        self.scrollingOffset = 0
        self.centeringState = None
        #print "clear. mousepos = ", point_to_string(self.lastMousePos)
        
    def onLineUpClicked(self):
        self.centeringState = None
        self.scrollUpLineWithoutUpdating()
        self.update()

    def scrollUpLineWithoutUpdating(self):
        if self.scrollingOffset != 0:
            self.scrollingOffset -= 1

    # BEGIN itemNPixelsAboveTheItem
    #
    # This function is a bit complicated, and deserves explanation.
    #
    #    itemNPixelsAboveTheItem(self, n, i) = j
    #
    # means
    #
    # "j is the maximum index of an item J such that: the distance
    # from J to I is greater than n. (where I is such that index(I) =
    # i)."
    #
    # We want to write this in a form that is as declarative as
    # possible, i.e. as close as possible to the above definition. To
    # do that, we first define the function distance(j, i):
    #
    def __itemNPixelsAboveTheItem(self, n, i):
        def distance(j, i):
            assert j <= i
            return reduce(operator.add,
                          [ item.height for item in self.items[j:i] ],
                          0)
        # now we apply the definition of j: j is the maximum of the
        # set of all indices in range(0,i) such that...
        distant_elements =  [ j for j in range(0,i) if  distance(j, i) >= n]
        if distant_elements ==[]:
            return None
        else:
            return max(distant_elements)

    # This is an imperative version by Thomas. --- update: Sorry
    # Thomas, this does not work: try scrolling page up to
    # verify. When I have time I'll try to see why. In the meantime,
    # let us use the declarative version. Unfortunately I had already
    # thrown away my old version when I saw yours.
    #
    # A: I accidently wrote ItemNPi... (big I), so it didnt work because
    # of that typo.
    def itemNPixelsAboveTheItem(self, n, i):
        deltaY = 0
        while i != 0:
            i -= 1
            deltaY += self.items[i].height
            if deltaY > n:
                return i + 1
        return None
    # END itemNPixelsAboveTheItem        

    def onPageUpClicked(self):
        self.centeringState = None
        self.scrollUpPageWithoutUpdating()
        self.update()
    def scrollDownLineWithoutUpdating(self):
        if self.scrollingOffset != len(self.items)-1:
            self.scrollingOffset += 1
    def scrollUpPageWithoutUpdating(self):
        if self.scrollingOffset > 0:
            i = self.itemNPixelsAboveTheItem(self.listArea.height()-1, self.scrollingOffset)
            if i is None:
                self.scrollingOffset = 0
            else:
                self.scrollingOffset = i
    def onLineDownClicked(self):
        self.centeringState = None
        self.scrollDownLineWithoutUpdating()
        self.update()
    def onPageDownClicked(self):
        self.centeringState = None
        x = self.itemAtPosition(self.listArea.height()-1)
        if not x is None:
            self.scrollingOffset = x
            self.listArea.paint(None)
        self.updateScrollButtonArrows()
    def onGotoTopClicked(self):
        self.centeringState = None
        if self.scrollingOffset != 0:
            self.scrollingOffset = 0
            self.update()
    def onGotoBottomClicked(self):
        self.centeringState = None
        self.scrollingOffset = len(self.items)-1
        self.scrollUpPageWithoutUpdating()
        self.scrollDownLineWithoutUpdating()
        self.update()
    def updateScrollButtonArrows(self):
        if self.theUpScrollArrowMustBeVisible():
            self.buttonUp.setPixmap(QPixmap(config.imgDir + "stock_up.png"))
            self.gotoTopBut.setPixmap(QPixmap(config.imgDir + "stock_top.png"))
            self.buttonBoxUp.setEnabled(True)
            self.gotoTopBut.setEnabled(True)
        else:
            self.buttonUp.setPixmap(QPixmap())
            self.gotoTopBut.setPixmap(QPixmap())
            self.buttonBoxUp.setEnabled(False)
            self.gotoTopBut.setEnabled(False)
        if self.theDownScrollArrowMustBeVisible():
            self.buttonDown.setPixmap(QPixmap(config.imgDir + "stock_down.png"))
            self.gotoBottomBut.setPixmap(QPixmap(config.imgDir + "stock_bottom.png"))
            self.buttonBoxDown.setEnabled(True)
            self.gotoBottomBut.setEnabled(True)
        else:
            self.buttonDown.setPixmap(QPixmap())
            self.buttonBoxDown.setEnabled(False)
            self.gotoBottomBut.setPixmap(QPixmap())
            self.gotoBottomBut.setEnabled(False)
    def theUpScrollArrowMustBeVisible(self):
        return len(self.items) > 0 and self.scrollingOffset > 0
    def theDownScrollArrowMustBeVisible(self):
        # @@ FIXME: bug if the last element of the list is
        # partially visible: you see half of it, and yet you
        # cannot scroll.
        v = len(self.items) > 0 and  self.lastVisibleItem() < len(self.items)-1
        #print "theDownScrollArrowMustBeVisible : ", v
        return v
      
    def lastVisibleItem(self):
        if len(self.items) > 0:
            x = self.itemAtPosition(self.listArea.height()-1)
            if x == None:
                return len(self.items) -1
            else:
                return x
        else:
            return None

    # this is more declarative (and slower) than the following. Feel
    # free to comment this and uncomment the imperative version.
##     def itemAtPosition(self, posy):
##         def distance(i, j, delta):
##             d1 = reduce( operator.add,
##                          [ item.height for item in self.items[i:j-1] ],
##                          0)
##             d2 = d1 + self.items[j].height
##             return delta in range(d1, d2)
        
##         return find( range(self.scrollingOffset, len(self.items)),
##                      lambda i: distance(self.scrollingOffset, i, posy))

    # the item at a given position can be None
    def itemAtPosition(self, posy):
        y = 0
        i = self.scrollingOffset
        while i < len(self.items) and y < self.listArea.height():
            item = self.items[i]
            if y + item.height > posy:
                return i
            else:
                y += item.height
                i += 1
        return None

    # call this after adding items, or you won't see the changes
    def update(self):
        #print "update called. mousePos = " , self.lastMousePos
        self.listArea.paint(self.lastMousePos)
        self.updateScrollButtonArrows()
    def yOffsetOfItem(self, item):
        # return the sum of the heights of all the items from
        # scrollingOffset to item :-)
        return reduce(operator.add ,
                      [i.height for i in self.items[self.scrollingOffset:self.items.index(item)]],
                      0)
    # changes self.scrollingOffset in order to show item in position
    # yoffset from the top (if possible)
    def centerItem(self, itemText, yoffset):
        #print "centering item  ", itemText, ' to y = ', yoffset
        item = find(self.items, lambda i:i.text == itemText)
        assert item != None
        self.centeringState = CenteringState(itemText, yoffset)
        # BEGIN calculate realYOffset for the item
        theItemWouldBeCompletelyInside = yoffset + item.height < self.listArea.height()
        if theItemWouldBeCompletelyInside:
            realYOffset = yoffset
        else:
            realYOffset = (self.listArea.height()-item.height) * 3/4
        # END

        # BEGIN calculate self.scrollingOffset

        # BEGIN mathematical definition (slow)
##         itemIdx = self.items.index(item)
##         i = find( range(itemIdx),
##                   lambda i: realYOffset >= sum( [x.height for x in self.items[i: itemIdx ]], 0))
##         if i != None:
##             self.scrollingOffset = i
##         else:
##             self.scrollingOffset = 0
        #END

        # BEGIN imperative version (fast)
        currentOffsetY = realYOffset
        i = self.items.index(item)
        while currentOffsetY > 0:
            if i > 0:
                i -= 1
                currentOffsetY -= self.items[i].height
            else:
                self.scrollingOffset = 0
                return
        self.scrollingOffset = i
        # END
        
        # END 


class WrappingButton(QHBox):
    def __init__(self, text, parent, data, toggle):
        QHBox.__init__(self, parent)
        self.text = text
        self.toggle_button = toggle
        self.pressed = False
        self.mouseOver = False
        self.setMargin(5)
        self.setBackgroundMode(Qt.NoBackground)
        self.recalculateHeight(self.width())
        self.data = data
    def setText(self, t):
        self.text = t
        self.recalculateHeight(self.width())
        self.repaint()
    def resizeEvent(self, ev):
        self.recalculateHeight(ev.size().width())
        self.repaint()
    def setPressedState(self, pressed):
        self.pressed = pressed
        self.repaint()
    def recalculateHeight(self, wt):
        f = QFont()
        qfontm = QFontMetrics(f)
        boundingRectangle = qfontm.boundingRect(0,0,
            wt - config.gWrappingButton_DistanceOfTextFromBorder*2,
            1000,
            Qt.BreakAnywhere,
            self.text)
        newHt = boundingRectangle.height() + config.gWrappingButton_DistanceOfTextFromBorder*2
        if newHt != self.height():
            self.setMinimumHeight(newHt)
    def mousePressEvent(self, mouseEvent):
        if self.toggle_button:
            self.pressed = not self.pressed
            if self.pressed:
                self.emit(PYSIGNAL("pressed"), (self,))
            else:
                self.emit(PYSIGNAL("released"), (self,))
        else:
                self.emit(PYSIGNAL("pressed"), (self,))
        self.repaint()
    def enterEvent(self, qevent):
        self.mouseOver = True
        self.repaint()
    def leaveEvent(self, qev):
        self.mouseOver = False
        self.repaint()
    def paintEvent(self, pevent):
        pix = QPixmap(self.size())
        painter = QPainter(pix)
        # BEGIN draw the button
        pix.fill(self, QPoint(0,0))
        r = QRect(0,0,self.width(), self.height())
        if self.isEnabled():
            lStyle = QStyle.Style_Enabled
            if self.pressed:
                lStyle |= QStyle.Style_Down
            else:
                lStyle |= QStyle.Style_Up
            if self.mouseOver:
                lStyle |= QStyle.Style_MouseOver
            self.style().drawPrimitive(QStyle.PE_ButtonCommand,
                           painter,
                           r,
                           self.colorGroup(),
                           lStyle)
        # END
        
        # BEGIN draw the text
        painter.setPen(self.palette().active().text())
        painter.drawText(config.gWrappingButton_DistanceOfTextFromBorder, config.gWrappingButton_DistanceOfTextFromBorder,
            self.width() - config.gWrappingButton_DistanceOfTextFromBorder*2,
            self.height() - config.gWrappingButton_DistanceOfTextFromBorder,
            Qt.BreakAnywhere, self.text)
        # END
        painter.end() # it is ok to do end before begin
        painter.begin(self)
        painter.drawPixmap(0,0,pix)


class MyLineEdit(QLineEdit):
    def __init__(self, parent):
        QLineEdit.__init__(self, parent)
        self.timerId = None
    def mousePressEvent(self, e):
        QLineEdit.selectAll(self)
    def timerEvent(self, ev):
        if ev.timerId() == self.timerId:
            self.emit(PYSIGNAL("timer"), (self,))
            # do not use killTimers here! QLineEdit uses its own timer
            # to manage the blinking cursor!
            self.killTimer(self.timerId)
            self.timerId = None
    def keyPressEvent(self, e):
        if self.timerId != None:
            # do not use killTimers here! QLineEdit uses its own timer
            # to manage the blinking cursor!
            self.killTimer(self.timerId)
        self.timerId = self.startTimer(1000)
        e.ignore()
        QLineEdit.keyPressEvent(self, e)

        
class MyComboBox(QComboBox):
    def __init__(self, parent):
        QComboBox.__init__(self, parent)
        QObject.connect(self, SIGNAL("activated(int)"), self.onActivated)
        self.currentId = 0
        self.idToItem = {}
    def insertItem(self,  icon, text, data):
        QComboBox.insertItem(self, QPixmap(icon), text, self.currentId)
        self.idToItem[self.currentId] = data
        self.currentId += 1
    def onActivated(self, i):
        self.emit(PYSIGNAL("itemActivated"), (self, self.idToItem[i])) 
    def clear(self):
        self.idToItem = {}
        self.currentId  = 0
        QPopupMenu.clear(self)

class PopupMenuItem:
    def __init__(self, name):
        self.name = name
    def stringForPopupMenu(self):
        return self.name

class MyPopupMenu(QPopupMenu):
    def __init__(self, parent):
        QPopupMenu.__init__(self, parent)
        self.currentId = 0
        self.fromIdToItem = {}
        QObject.connect(self, SIGNAL("activated(int)"), self.onActivated)
    def addItem(self, item, icon, get_string):
        self.insertItem(QIconSet(icon), get_string(item), self.currentId)
        self.fromIdToItem[self.currentId] = item
        self.currentId += 1
    def onActivated(self, id):
        item = self.fromIdToItem[id]
        self.emit(PYSIGNAL("itemActivated"), (self, item))
    def clear(self):
        self.fromIdToItem = {}
        self.currentId  = 0
        QPopupMenu.clear(self)

