#!/usr/bin/env python
# -*- coding: iso8859-1 -*-
"""
   OneFinger - main module

   @license: GNU GPL, see COPYING for details
   @copyright: 2004, Maurizio Colucci
"""

import sys, os,  datetime, pickle, shutil, config, persistence, string
#import kdecore, kdeui, config

from sets import Set
from classes import LocalFileSystemDirectory
from util import createDirIfNotExisting, find, addSlashIfNeeded, ext, first, forall, remove_ending_slash_if_not_root
from qt import QMessageBox, QPixmap, QIconSet, QObject, PYSIGNAL, QToolBar, SIGNAL, QVBox
from qt import QHBox, QMainWindow, QApplication, QString, QTextEdit
from qt import QFrame, Qt, QSizePolicy, QPushButton, QGrid, QLabel, QToolTip
from my_qt_widgets import FlashingButton, MyPopupMenu, FlashingLabel, WrappingButton, MyListBox, MyLineEdit, MyComboBox, TextItem, TextSeparator, PopupMenuItem, FlashingTextEdit, text_item_button_info
def _(x):
    return x





listBox = {}
sortBox = {}

filterBox = {}
filterButton = {}
filterMenu = {}
filterQuickText = {}
filterClearButton = {}

hBox = {}
icon = {}

last_clicked_program = None

files_whose_light_bulb_is_pressed = []

message_you_can_now_select_a_file = _("You can now execute the command above, or click additional files or arguments.")
def point_to_string(p):
    if p is None:
        return None
    else:
        return p.x, p.y


class WindowInfo:
    def __init__(self, wid, name):
        self.wid = wid
        self.name = name
    def stringForPopupMenu(self):
        return self.name

#class MyMainWindow(kdeui.KMainWindow):
class MyMainWindow(QMainWindow):
    
    # BEGIN used on qt. 
    def closeEvent(self, ev):
        global recent_dirs, visibleLocationInFilePanel
        recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
        QMainWindow.closeEvent(self, ev)
        # BEGIN removed because I call closeEvent
        #return True # accept the closing
        # END
    # END
    
    # BEGIN used by kde, useless in the qt version
    #    def queryClose(self):
    #        global recent_dirs
    #        recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
    #        return True # accept the closing
    # END

def clearFileFiltersAndUpdate():
    clearAux('file')
def clearArgumentFiltersAndUpdate():
    clearAux('argument')
def clearProgramFiltersAndUpdate():
    clearAux('program')

def clearAux(k):
    filterQuickText[k].clear()
    redraw_programs()
    redraw_files(centering = True)
    redraw_arguments()

class argument_info:
    def __init__(self, name, usage_count, last_usage_time):
        self.name = name
        self.usage_count = usage_count
        self.last_usage_time = last_usage_time
        
def redraw_arguments():
    def split_and_sort_args(pairs, program):
        k = 'argument'
        tx = sortBox[k].currentText()
        if tx == _("Name"):
            pairs.sort(lambda p1, p2: cmp(p1.name, p2.name))
            names = [x.name for x in pairs]
            return [(_('Sorted by name (used with ') + program + ")", names)]
        elif tx == _('Recent usage'):
            pairs.sort(lambda p1, p2: -cmp(p1.last_usage_time, p2.last_usage_time))
            names = [x.name for x in pairs]
            return [  (_('Recently used with ') + program, names)]
        elif tx == _('Frequent usage'):
            pairs.sort(lambda p1, p2: -cmp(p1.usage_count, p2.usage_count))
            names = [x.name for x in pairs]
            return [  (_('Frequently used with ') + program, names)]
        else:
            raise "error"
    k = 'argument'
    listBox[k].clear()
    if last_clicked_program != None:
        pairs = arguments_of_program.get(last_clicked_program , [])
        filt = str(filterQuickText[k].text())
        pairs_filtered = [a for a in pairs if  is_substring_of(filt, a.name)]
        l = split_and_sort_args(pairs_filtered, last_clicked_program)
        for i in l:
            (title, args) = i
            listBox[k].items.append(TextSeparator(title
                                                  , listBox[k].listArea.width()
                                                  , scalable))
            for x in args:
                # BEGIN create icon
                if os.path.isdir(x):
                    ic = QPixmap(config.imgDir + "folder.png") 
                elif os.path.isfile(x):
                    ex = ext(x)
                    if ex.endswith('~'):
                        ex2 = ex[1:-1]
                    else:
                        ex2 = ex[1:]
                    ic =  QPixmap(config.imgDir + typeToIcon.get(ex2, "generic.png"))
                else:
                    ic = QPixmap()
                # END
                listBox[k].items.append(TextItem(
                    text = x,
                    icon = ic,
                    buttons_info = [],
                    isSelectable =  main_program_in_command(str(commandEntry.text())) != None,
                    width = listBox[k].listArea.width(),
                    pressed = False,
                    data = x,
                    scalable_values = scalable))
            if args == []:
                listBox[k].items.append(TextItem(text = "(no argument used so far)",
                                                 icon = QPixmap(),
                                                 buttons_info = [],
                                                 isSelectable = False,
                                                 width = listBox[k].listArea.width(),
                                                 pressed = False,
                                                 data = "",
                                                 scalable_values = scalable))

    listBox[k].update()

def onFilterKeyTimer(widget):
    applyFilter()
    
def onEnterPressedOnFilter():
    applyFilter()
    
def applyFilter():
    global recent_filters
    redraw_programs()
    redraw_files()
    redraw_arguments()

    # BEGIN rebuild recent_filters[k] for all k
    for k in ['program', 'file', 'argument']:
        t = str(filterQuickText[k].text())
        if t:
            if t in recent_filters[k]:
                recent_filters[k].remove(t)
            recent_filters[k] = [t] + recent_filters[k]
        recent_filters[k] = (recent_filters[k])[:config.gMaxItemsInRecentLocMenu] # truncate
    # END
    rebuildFilterMenus()

def rebuildFilterMenus():
    for k in ['program', 'file', 'argument']:
        filterMenu[k].clear()
        l = recent_filters[k]
        l.reverse()
        for f in l:
            filterMenu[k].addItem(f, QPixmap(), lambda x:x)
        l.reverse()

def onRecentFilterClicked(menu, filter):
    print "onRecentFilterClicked"
    k = find(['argument', 'file', 'program'], lambda x: menu is filterMenu[x])
    filterQuickText[k].setText(filter)
    applyFilter()
         

              
def split_and_sort_files_or_progs(k, objs2):
    if k == 'file':
        last_usage = last_usage_time_of_dir
        usage_count = usage_count_of_dir
    elif k == 'program':
        last_usage = last_usage_time_of_program
        usage_count = usage_count_of_program
    tx = sortBox[k].currentText()
    if  tx == _("Name"):
        objs2.sort(lambda o1, o2: cmp(o1.lower(), o2.lower()))
        return [(_('Sorted by name'), objs2)]
    elif tx == _('Recent usage'):
        usedObjs = [ x for x in objs2 if last_usage.get(x, None) != None]
        unusedObjs = [x for x in objs2 if last_usage.get(x, None) == None]
        unusedObjs.sort(lambda o1, o2: cmp(o1.lower(),   o2.lower()))
        usedObjs.sort(lambda o1, o2: -cmp(last_usage[o1],last_usage[o2]))
        return [  (_('Recently used'), usedObjs) ,
                  (_('Never used'), unusedObjs)]
    elif tx == _("Frequent usage"):
        usedObjs = [x for x in objs2 if usage_count.get(x, 0) != 0]
        unusedObjs = [x for x in objs2 if usage_count.get(x, 0) == 0]
        unusedObjs.sort(lambda o1, o2: cmp(o1.lower(), o2.lower()))
        usedObjs.sort(lambda o1, o2: - cmp(usage_count[o1], usage_count[o2]))
        return [
                (_('Frequently used'), usedObjs),
                (_('Never used'), unusedObjs)]
    elif tx  == _("Date"):
        objs2.sort(lambda o1, o2: -cmp(os.path.getmtime(o1), os.path.getmtime(o2)))
        return [ (_('Most recently modified first'), objs2)]
    elif tx == _("Size"):
        withSize = [o for o in objs2 if os.path.isfile(o)]
        withoutSize = [ o for o in objs2 if not os.path.isfile(o)]
        withSize.sort(lambda o1, o2: -cmp(os.path.getsize(o1), os.path.getsize(o2)))
        withoutSize.sort(lambda o1, o2: cmp(o1.lower(), o2.lower()))
        return [ (_('Biggest first'), withSize),
                 (_('No size'), withoutSize)]
    else:
        raise "error: the current text is invalid: " + tx + " " + sortBox[k].currentText()

def onLocationButtonPressed():
    #command = str(commandEntry.text()) + cur_dir + " "
    commandEntry.insert(cur_dir + "  ")
    commandEntry.setFocus()
    #commandEntry.moveCursor(QTextEdit.MoveEnd, False)


def redraw_files(centering = False):
    if centering:
        offs = listBox['file'].scrollingOffset
    locationButton.setText(cur_dir)
    locationButton.data = cur_dir
    locationButton.setEnabled( main_program_in_command(str(commandEntry.text())) != None)
    set_current_dir_button.setOn(cur_dir == os.getcwd())
    narrow_dir_button.setOn(cur_dir in files_whose_light_bulb_is_pressed)
    listBox['file'].clear()
    cur_dir_slash = addSlashIfNeeded(cur_dir)
    files_without_path = [x for x in os.listdir(cur_dir)]
    files_with_path = [ cur_dir_slash + x for x in files_without_path]
    dirs = [o for o in files_with_path if os.path.isdir(o)]
    ordinary_files = [o for o in files_with_path if os.path.isfile(o)]
    executable_files = [o for o in ordinary_files if
                        os.access(o, os.X_OK)]
    filt = str(filterQuickText['file'].text())
    interesting_files = [x for x in dirs + ordinary_files if is_substring_of(filt, x)]
    l = split_and_sort_files_or_progs('file', interesting_files)
    for i in l:
        (title, files) = i
        if files != []:
            listBox['file'].items.append(TextSeparator(title
                                                       , listBox[k].listArea.width()
                                                       , scalable))
        for x in files:
            #BEGIN make icons (ic =icon, bic = button icon)
            if x in files_whose_light_bulb_is_pressed:
                press = "pressed"
            else:
                press = "released"
            if x in dirs:
                ic = QPixmap(config.imgDir +  "folder.png")
                bic = [ text_item_button_info( [ QPixmap(config.imgDir +  "enter-no-arrow.png")
                                                 , QPixmap(config.imgDir + "enter.png")]
                                               , "enter directory"
                                               , "not a pushbutton" )
                        , text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                                 , QPixmap(config.imgDir + "lit-bulb-small.png")]
                                                , "narrow", press)]
            else:
                ex = ext(x)
                if ex.endswith('~'):
                    ex2 = ex[1:-1]
                else:
                    ex2 = ex[1:]
                ic =  QPixmap(config.imgDir + typeToIcon.get(ex2, "generic.png"))
                
                bic = [text_item_button_info([QPixmap(config.imgDir + "bulb.png")
                                              ,QPixmap(config.imgDir + "lit-bulb-small.png")]
                                             , "narrow", press)]
            # END
            listBox['file'].items.append(TextItem(
                text = os.path.basename(x),
                icon = ic,
                buttons_info = bic,
                isSelectable =  ( main_program_in_command(str(commandEntry.text())) != None
                                  or x in executable_files),
                width = listBox['file'].listArea.width(),
                pressed = False,
                data = x,
                scalable_values = scalable))
    # BEGIN scroll the list to restore the previous scrolling position
    if centering:
        listBox['file'].scrollingOffset = offs
#        (text, offset) = centering
#        listBox['file'].centerItem(text, offset)
    #END
    listBox['file'].update()

def is_substring_of(s, st):
    return st.lower().find(s.lower()) != -1


program_icons = {

    "xpdf": "xpdf.png"
    , "gnome-terminal": "gnome-terminal.png"
    , "konsole": "konsole.png"
    , "aterm": "konsole.png"
    , "none": "italian-flag.png"
    , "kfind": "kfind.png"
    , "kedit": "kwrite.png"
    , "kate": "kwrite.png"
    , "kbear": "kbear.png"
    , "gedit": "gedit.png"
    , "kwrite": "kwrite.png"
    , "krename": "name.png"
    , "kmix": "kmix.png"
    , "aumix": "aumix.png"
    , "ksnapshot": "ksnapshot.png"
    , "gnome-screenshot": "gnome-screenshot.png"
    , "kcontrol": "kcontrol.png"
    , "kparted": "devices.png"
    , "qtparted": "devices.png"
    , "diskdrake": "diskdrake.png"
    , "kdf": "kdiskfree.png"
    , "pan": "pan.png"
    , "knews": "news.png"
    , "dvdrip": "cdrom.png"
    , "kgpg": "kgpg.png"
    , "ark": "ark.png"
    , "kuickshow": "kuickshow.png"
    , "eog": "eog.png"
    , "gthumb": "gthumb.png"
    , "korganizer": "korganizer.png"
    , "sylpheed": "sylpheed.png"
    , "abiword": "abiword.png"
    , "nautilus": "nautilus.png"
    , "gftp": "gftp.png"
    , "kpdf": "generic.png"
    , "kghostview": "kghostview.png"
    , "kdvi": "kdvi.png"
    , "gpdf":"gpdf.png"
    , "gimp": "gimp.png"
    , "k3b": "k3b.png"
    , "emacs": "emacs.png"
    , "cervisia": "cervisia.png"
    , "mplayer": "mplayer.png"
    , "oowriter": "ooowriter.png"
    , "texmacs": "texmacs.png"
    , "konqueror": "konqueror.png"
    , "firefox": "firefox.png"
    , "mozilla-thunderbird": "mozicon16.xpm"
    , "mozilla": "mozilla.png"
    , "opera": "opera.png"
    , "evolution": "evolution.png"
    , "kmail": "kmail.png"
    , "knode": "knode.png"
    , "amule": "amule.png"
    , "synaptic": "synaptic.png"
    , "gmplayer": "mplayer.png"
    , "umbrello": "umbrello.png"
    , "xine": "xine.png"
    , "xchat": "xchat.png"
    , "xcalc": "xcalc.png"
    , "file-roller": "fileroller.png"
    , "totem": "totem.png"
    , "xmms": "xmms.png"
    , "kaffeine": "kaffeine.png"
    , "gxine": "xine.png"
    }

def extensions_selected_for_narrowing():
    s = Set()
    for f in files_whose_light_bulb_is_pressed:
        if os.path.isdir(f):
            s.add("<directory>")
        else:
            e = ext(f)
            s.add(e)
    return s

def redraw_programs():
    def set_to_string(s):
        l = []
        for x in s:
            l.append( x)
        st = "\""
        for x in range(0, len(l)):
            if l[x] == "":
                l[x] = "<no extension>"
            st += l[x]
            if x<len(l)-1:
                st += ", "
        return st + "\""
    k = 'program'
    listBox[k].clear()
    filt = str(filterQuickText[k].text())
    filtered_progs = [p for p in programs if is_substring_of(filt, p)
                      if forall(extensions_selected_for_narrowing(),
                                lambda e: e in types_manageable_by_program.get(p, Set()))]
    l = split_and_sort_files_or_progs(k, filtered_progs)
    trunc = 100
    for i in l:
        (sorting_string, progs) = i
        if progs != [] or len(extensions_selected_for_narrowing())> 0:
            if filt != "":
                filter_string =  _("Containing \"") + filt + "\"; "
            else:
                filter_string = ""
            if len(extensions_selected_for_narrowing())>0:
                narrowing_string = ( _("Can manage ")
                                     + set_to_string(extensions_selected_for_narrowing())
                                     + "; ")
            else:
                narrowing_string = ""
            listBox[k].items.append(TextSeparator(narrowing_string
                                                  +filter_string 
                                                  + sorting_string
                                                  , listBox[k].listArea.width()
                                                  , scalable))
        for x in progs[:trunc]:
            ic =  QPixmap(config.imgDir + program_icons.get(x, "generic.png"))
            
            listBox[k].items.append(TextItem(
                text = x,
                icon = ic,
                buttons_info = [text_item_button_info([QPixmap(config.imgDir + "glasses.png")
                                                       , QPixmap(config.imgDir + "glasses.png")]
                                                      , "show arguments"
                                                      , "not a pushbutton")],
                isSelectable = True,
                width = listBox[k].listArea.width(),
                pressed = False,
                data = x,
                scalable_values = scalable))
        if len(progs) > trunc:
            listBox[k].items.append(TextItem(
                text = _("(The list was too long and has been truncated. Use the filter box below to find the program you need.)"),
                icon = QPixmap(),
                buttons_info = [],
                isSelectable = False,
                width = listBox[k].listArea.width(),
                pressed = False,
                data = "",
                scalable_values = scalable))
            
    listBox[k].update()

typeToIcon = {
    'txt': 'text-file.png',
    'ics': 'schedule.png',
    'pdf': 'pdf.png',
    'dvi': 'kdvi.png',
    'ps':  'kdvi.png',
    'doc': 'word-26x26.png',
    'rpm': 'rpm.png',
    'mp3': 'audio.png',
    'flac': 'audio.png',
    'ogg': 'audio.png',
    'wav': 'audio.png',
    'mpc': 'audio.png',
    'flac': 'audio.png',
    'ogm': 'video.png',
    'avi': 'video.png',
    'mpg': 'video.png',
    'mpeg': 'video.png',
    'sxw': 'ooowriter.png',
    'wma': 'audio.png',
    'wmv': 'video.png',
    'spec': 'rpm.png',
    'asf': 'video.png',
    'zip': 'archive.png',
    'rar': 'archive.png',
    'tgz': 'archive.png',
    'gz':  'archive.png',
    'bz2': 'archive.png',
    'tar': 'archive.png',
    'arj': 'archive.png',
    'ace': 'archive.png',
    'htm': 'internet.png',
    'html': 'html.png',
    'png': 'picture.png',
    'gif': 'picture.png',
    'jpg': 'picture.png',
    'ppm': 'picture.png',
    'bmp': 'picture.png',
    'xpm': 'picture.png',
    'blend': 'picture.png',
}



def add_cur_dir_to_recent_dirs(recent_dirs):
    r = find(recent_dirs, lambda x: x== cur_dir)
    if not r is None:
        recent_dirs.remove(cur_dir)
    recent_dirs2 = [cur_dir] + recent_dirs
    rebuild_recent_locations_menu(recent_dirs2)
    return recent_dirs2[:config.gMaxItemsInRecentLocMenu]
  


def rebuild_recent_locations_menu(recent_dirs):
    recentLocPopupMenu.clear()
    for l in recent_dirs:
        l2 = remove_ending_slash_if_not_root(l)
        ic = QPixmap(config.imgDir + "folder.png")
        recentLocPopupMenu.addItem(l2, ic, lambda x:x)

def rebuild_important_directories_menu():
    favouriteLocPopupMenu.clear()
    for l in important_directories:
        l2 = remove_ending_slash_if_not_root(l)
        ic = QPixmap(config.imgDir + "folder.png")
        favouriteLocPopupMenu.addItem(l2, ic, lambda x:x)
        
# BEGIN Qt only
applic = QApplication(sys.argv)
# END

real_resolution = QApplication.desktop().width()

resolution_multiplier = 1# real_resolution/1024.0


class scalable_values:
    def __init__(self, resolution_multiplier):
        self.resolution_multiplier = resolution_multiplier
        self.gListBoxMarginOfFlashingVBox = 5 * resolution_multiplier
        self.gListBoxMarginOfNonFlashingVBox = 1 * resolution_multiplier

        self.gWrappingButton_DistanceOfTextFromBorder = 8 * resolution_multiplier
        self.gSpaceAboveTextSeparator = 13 * resolution_multiplier
        self.gSpaceBelowTextSeparator = 11 * resolution_multiplier
        print "self.gSpaceBelowTextSeparator " + str(self.gSpaceBelowTextSeparator)
        print "self.gSpaceAboveTextSeparator " + str(self.gSpaceAboveTextSeparator)


scalable = scalable_values(resolution_multiplier)

# BEGIN KDE only
# this must be before the creation of MyMainWindow
#applic = kdecore.KApplication(sys.argv, "OneFinger")
# END

createDirIfNotExisting(config.dataDir)

def rescan_programs():
    global programs
    programs = []
    pathlist = os.environ['PATH'].split(os.pathsep)
    for path in pathlist:
        # find prgs really there and allowed to execute
        if os.path.isdir(path): # covers the case where the user has an unexisting path in $PATH
            for p in os.listdir(path):
                if os.access(os.path.join(path,p), os.X_OK):
                    programs.append(p)

rescan_programs()


# BEGIN build default arguments for programs
default_arguments = [ ("mplayer", [ "-vo", "sdl", "xv", "--enqueue", "-sub"])
                      ,("gmplayer", [ "-vo", "sdl", "xv", "--enqueue"])
                      ,("totem", [ "--enqueue"])
                      ,("konsole", [ "--workdir"])
                      ,("xmms", [ "--enqueue"])
                      ,("k3b", [ "--emovixdvd", "--copycd", "--clonecd", "--erasecd", "--isoimage", "--binimage",
                                 "--formatdvd", "--datacd", "--audiocd", "--videocd", "--mixedcd", "--datadvd" ])
                      ,("xine", [ "--enqueue"])
                      ,("ark", [ "--extract", "--extract-to", "--add", "--add-to", "--guess-name"])
                      ,("file-roller", [ "--extract", "--extract-to=", "--add", "--add-to="])
                      ,("cp", [ "-r"])
                      ,("scp", [ "-r"])
                      ,("dpkg", [ "-c","-i", "--install",  "--contents"])
                      ,("mv", [ "-r"])
                      ,("du", [ "-s", "-h", "-sh", "--human-readable", "--summarize"])
                      ,("rm", [ "-r", "-f", "-rf"])
                      ,("btdownloadcurses", [ "--url"])
                      ,("btdownloadgui", [ "--url"])
                      ,("df", [ "--human-readable"])
                      ,("ls", [ "--color", "-a", "-l", "-la"])
                      ,("emacs", [ "-nw"])
                      ,("echo", [ "$PATH"])
                      ,("apt-get", [ "update", "upgrade", "clean", "autoclean"])
                      ,("update-alternatives", [ "--config", "x-window-manager", "x-terminal-emulator"
                                                 ,"x-www-browser", "x-session-manager", "editor"
                                                 , "emacs", "c++", "cc", "cli", "ctags", "editor", "dia"])
                      ]
default_arguments2 = {}
for pair in default_arguments:
    (prog_name, arg_list) = pair
    l2 = [argument_info(x, 1, datetime.datetime.now()) for x in arg_list]
    default_arguments2[prog_name] = l2

# END
                      
# BEGIN read database
help_bar_visible =persistence.loadVariableFromFile(config.dataDir + 'help_bar_visible'
                                                   , default = True)
types_manageable_by_program = persistence.loadVariableFromFile(
    config.dataDir + 'types_manageable_by_program', default = {})
recent_dirs = persistence.loadVariableFromFile(config.dataDir + 'recent_dirs', default = [])
program_runs_in_terminal = persistence.loadVariableFromFile(config.dataDir + 'program_runs_in_terminal', default = {})
cur_dir = persistence.loadVariableFromFile(config.dataDir + 'cur_dir', default = '/')
arguments_of_program = persistence.loadVariableFromFile(
    config.dataDir + 'arguments_of_program',
    default = default_arguments2)
important_directories = persistence.loadVariableFromFile(config.dataDir
                                                         + 'important_directories'
                                                         , default = ['/'])
last_usage_time_of_dir = persistence.loadVariableFromFile(config.dataDir + 'last_usage_time_of_dir', default = {})
last_usage_time_of_program = persistence.loadVariableFromFile(
    config.dataDir + 'last_usage_time_of_program', default = {})
recent_filters = persistence.loadVariableFromFile(config.dataDir + 'recent_filters',
                                                  default = {'file': [], 'argument':[], 'program':[]})
sorting_criterion_for_dir = persistence.loadVariableFromFile(config.dataDir + 'sorting_criterion_for_dir', default = {})
sorting_criterion_for_arguments = persistence.loadVariableFromFile(
    config.dataDir + 'sorting_criterion_for_arguments'
    , default = "Recent usage")
sorting_criterion_for_programs = persistence.loadVariableFromFile(
    config.dataDir + 'sorting_criterion_for_programs'
    , default = "Frequent usage")

usage_count_of_dir = persistence.loadVariableFromFile(config.dataDir + 'usage_count_of_dir', default = {})
usage_count_of_program = persistence.loadVariableFromFile(config.dataDir + 'usage_count_of_program', default = {})
# END


mainWindow = MyMainWindow()
mainWindow.setCaption("OneFinger - the graphical shell")
mainWindow.setIcon(QPixmap(config.imgDir + "italian-flag.png"))
applic.setMainWidget(mainWindow)



#toolbar = kdeui.KToolBar(mainWindow)
toolbar = QToolBar(mainWindow)


bookmarksButton = FlashingButton(QIconSet(QPixmap(config.imgDir +"gnome-favorites.png")),
                 _("Important directories"),
                 toolbar)
bookmarksButton.setMinimumHeight(50 * resolution_multiplier)
QToolTip.add(bookmarksButton,

 _("Use this to change directory. The contents of the\ndirectory will appear in the panel below and to the left. \nIn order to add/remove a directory to/from this list\nyou can use the toolbar buttons to the right."))


favouriteLocPopupMenu = MyPopupMenu(bookmarksButton)



def onLocationSelectedFromPopupMenu(menu, new_dir):
    global recent_dirs, usage_count_of_dir, cur_dir
    recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
    cur_dir = new_dir
    filterQuickText['file'].clear()
    usage_count_of_dir[new_dir] = usage_count_of_dir.get(new_dir, 0) +1
    last_usage_time_of_dir[new_dir] = datetime.datetime.now()
    rebuild_recent_locations_menu(recent_dirs)
    cr = _(sorting_criterion_for_dir.get(new_dir, "Recent usage")) # sic!
    sortBox['file'].setCurrentText(cr)
    redraw_files()
    listBox['file'].startFlashing(config.gFlashingTimeForList,
                                  resetOriginalColorsWhenFinished = True)


QObject.connect(favouriteLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)
bookmarksButton.setPopup(favouriteLocPopupMenu)
bookmarksButton.show()

recent_dirsButton = FlashingButton(QIconSet(QPixmap(config.imgDir +"clock.png")),
                                       _("Recent directories"), toolbar)
recent_dirsButton.setMinimumHeight(50* resolution_multiplier)
recent_dirsButton.show()
QToolTip.add(recent_dirsButton,  _("Use this to change directory. The contents of the\ndirectory will appear in the panel below and to the left."))

recentLocPopupMenu = MyPopupMenu(recent_dirsButton)
recent_dirsButton.setPopup(recentLocPopupMenu)
QObject.connect(recentLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)


refresh_files_button = QPushButton(QIconSet(QPixmap(config.imgDir + "refresh.png")),
                                   _("Refresh files"),
                                   toolbar)
refresh_files_button.setMinimumHeight(50 * resolution_multiplier)
refresh_files_button.show()
QToolTip.add(refresh_files_button,
             _("Click here to update the file list (useful in case some external program modified the files on the disk)."))

def on_refresh_files_pressed():
    redraw_files()
    listBox['file'].startFlashing(config.gFlashingTimeForList,
                                  resetOriginalColorsWhenFinished = True)
    
QObject.connect(refresh_files_button, SIGNAL("pressed()"), on_refresh_files_pressed)

rescan_programs_buttons = QPushButton(QIconSet(QPixmap(config.imgDir + "refresh.png")),
                                   _("Rescan programs"),
                                   toolbar)
rescan_programs_buttons.setMinimumHeight(50* resolution_multiplier)
rescan_programs_buttons.show()
QToolTip.add(rescan_programs_buttons,
             _("Click here when you have installed or removed a program, or if the $PATH has changed. \nIt rescans the $PATH."))
def on_rescan_programs_pressed():
    rescan_programs()
    redraw_programs()
    listBox['program'].startFlashing(config.gFlashingTimeForList,
                                     resetOriginalColorsWhenFinished = True)

QObject.connect(rescan_programs_buttons, SIGNAL("pressed()"), on_rescan_programs_pressed)



add_to_important_dirs_button = QPushButton(QIconSet(QPixmap(config.imgDir + "gnome-favorites.png"))
                                           , "", toolbar)
add_to_important_dirs_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(add_to_important_dirs_button
             , _("Add the directory that is currently visible in the left panel to the important directories."))
def on_add_to_important_dir_clicked():
    print "here"
    global important_directories
    if not (cur_dir in important_directories):
        important_directories.append(cur_dir)
    rebuild_important_directories_menu()
    bookmarksButton.startFlashing(config.gFlashingTimeForButtons,
                                  resetOriginalColorsWhenFinished = True)

QObject.connect(add_to_important_dirs_button, SIGNAL("clicked()")
                , on_add_to_important_dir_clicked)

remove_from_important_dirs_button = QPushButton(QIconSet(QPixmap(config.imgDir + "remove-from-favorites.png"))
                                                , "", toolbar)
remove_from_important_dirs_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(remove_from_important_dirs_button
             , _("Remove the directory that is currently visible in the left panel from the important directories."))
def on_remove_from_important_dir_clicked():
    global important_directories
    if cur_dir in important_directories:
        if cur_dir == "/":
            QMessageBox.critical(mainVBox, _("Error"),
                                 _("The / directory cannot be removed."))
        else:
            important_directories.remove(cur_dir)
            rebuild_important_directories_menu()
            bookmarksButton.startFlashing(config.gFlashingTimeForButtons,
                                          resetOriginalColorsWhenFinished = True)
QObject.connect(remove_from_important_dirs_button, SIGNAL("clicked()")
                , on_remove_from_important_dir_clicked)



show_help_button = FlashingButton(QIconSet(QPixmap(config.imgDir + "help.png"))
                                  , "", toolbar)
show_help_button.setMinimumHeight(50* resolution_multiplier)
QToolTip.add(show_help_button, _("Restores the help bar."))
if  help_bar_visible:
    show_help_button.hide()
else:
    show_help_button.show()
def on_show_help_clicked():
    help_hbox.show()
    show_help_button.hide()
    global help_bar_visible
    help_bar_visible = True
QObject.connect(show_help_button, SIGNAL("clicked()"), on_show_help_clicked)

spacing = 4 * resolution_multiplier

mainVBox = QVBox(mainWindow)
mainVBox.show()
mainVBox.setMargin(spacing)
mainVBox.setSpacing(spacing)
mainWindow.setCentralWidget(mainVBox)


command_hbox = QHBox(mainVBox)
command_hbox.setSpacing(spacing)
clear_command_button = QPushButton( command_hbox)
clear_command_button.setPixmap(QPixmap(config.imgDir + "delete-right-big.png"))
clear_command_button.setMinimumSize(44 * resolution_multiplier
                                    ,32 * resolution_multiplier)
QToolTip.add(clear_command_button,
             _("Clears the command area (to the right)."))

def on_clear_command_button_pressed():
    global last_clicked_program
    commandEntry.setText("")
    commandEntry.setFamily("mono") # seems necessary, because it resets itself
    last_clicked_program = None
    redraw_arguments()
    # BEGIN We must not do this: it is the job of on_text_changed to
    # refresh (and flash) the file listbox. 
    #listBox['file'].enable_all_items(False)
    #listBox['file'].update()
    #locationButton.setEnabled(False)
    #END
    

QObject.connect(clear_command_button, SIGNAL("pressed()"), on_clear_command_button_pressed)
clear_command_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

commandEntry = FlashingTextEdit(command_hbox)
commandEntry.setFamily("mono")
commandEntry.setTextFormat(Qt.PlainText)
#commandEntry.setWrapPolicy(QTextEdit.AtWordBoundary)
#commandEntry.setWrapPolicy(QTextEdit.AtWordOrDocumentBoundary)
commandEntry.setWrapPolicy(QTextEdit.Anywhere)
commandEntry.setWordWrap(QTextEdit.WidgetWidth)
commandEntry.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
commandEntry.setMaximumHeight(50* resolution_multiplier)
def on_return_pressed_on_command():
    # BEGIN delete the last character, the newline
    commandEntry.moveCursor(QTextEdit.MoveBackward, True)
    commandEntry.delChar()
    # END
    execute(use_terminal = True)
QObject.connect(commandEntry, SIGNAL("returnPressed()"), on_return_pressed_on_command)
def on_text_changed():
    global previous_text
    cur_text = str(commandEntry.text())
    cur_p = main_program_in_command(cur_text)
    prev_p = main_program_in_command(previous_text)
    if  cur_p != None and prev_p == None:
        listBox['file'].startFlashing(config.gFlashingTimeForList,
                                      resetOriginalColorsWhenFinished = True)
        # BEGIN optimization: instead of redraw_files(centering=True),
        # which was too slow, we call:
        listBox['file'].enable_all_items(True)
        listBox['file'].update()
        locationButton.setEnabled(True)
        # END
        help_label.setText(message_you_can_now_select_a_file)
        help_label.stopFlashing(resetColor = True)
        program_label.stopFlashing(resetColor = True)


        #BEGIN set proper flashing for execute buttons
        terminal = program_runs_in_terminal.get(cur_p, None)
        if terminal == "yes":
            execute_button.stopFlashing()
            execute_in_term_button.startFlashing(config.gFlashingTimeForExecuteButtons,
                                                 resetOriginalColorsWhenFinished = False)
        elif terminal == "no":
            execute_in_term_button.stopFlashing()
            execute_button.startFlashing(config.gFlashingTimeForExecuteButtons,
                                         resetOriginalColorsWhenFinished = False)
        else:
            execute_in_term_button.stopFlashing()
            execute_button.stopFlashing()
        # END
    if  cur_p == None and prev_p != None:
        # BEGIN optimization: instead of redraw_files(centering=True),
        # which was too slow, we call:
        cur_dir_slash = addSlashIfNeeded(cur_dir)
        files_with_path = [cur_dir_slash+x for x in os.listdir(cur_dir)]
        executable_files = [o for o in files_with_path if os.path.isfile(o)
                          if  os.access(o, os.X_OK)]
        listBox['file'].enable_these_items(executable_files)
        listBox['file'].update()
        locationButton.setEnabled(False)
        # END
        help_label.setText(initial_message)
        help_label.startFlashing(config.gFlashingTimeForHelpMessage
                                 , resetOriginalColorsWhenFinished = True)
        program_label.startFlashing(config.gFlashingTimeForHelpMessage
                                    , resetOriginalColorsWhenFinished = True)

        execute_button.stopFlashing()
        execute_in_term_button.stopFlashing()
    previous_text = cur_text
    
QObject.connect(commandEntry, SIGNAL("textChanged()"), on_text_changed)
commandEntry.show()
commandEntry.setFocus()

QToolTip.add(commandEntry,
             
_("This is the command area. The programs, files and arguments \nyou click will appear here. When you have composed a meaningful\ncommand, you can execute it via the buttons on the right.\nYou can also type text directly, but this is only useful \nfor typing short program names, http locations, etc. \nYou should never type the same string more than once."))



previous_text = ""

exec_pix = QPixmap(config.imgDir + "action.png")
execute_button = FlashingButton(QIconSet(exec_pix),
                                _("Execute"),
                                command_hbox)
execute_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
execute_button.show()
execute_button.setMinimumHeight(exec_pix.height() + 5 * resolution_multiplier)
QToolTip.add(execute_button,
             _("Execute the command contained in the area to the left."))
exec_in_term_pix = QPixmap(config.imgDir + "gnome-terminal.png")
execute_in_term_button = FlashingButton(QIconSet(exec_in_term_pix),
                                        _("Execute \nin terminal"),
                                        command_hbox)
execute_in_term_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
execute_in_term_button.show()
execute_in_term_button.setMinimumHeight(exec_in_term_pix.height() + 5* resolution_multiplier)
QToolTip.add(execute_in_term_button,

_("Open a terminal and execute the command into it. Useful if the program\nproduces an output that you want to see. In order to configure which\nterminal to use, edit the file config.py. The default is konsole."))

def main_program_in_command(cmd):
    if len(cmd)>0:
        l = split_spaces_but_not_backslash_space(cmd)
        if len(l)>0:
            p = string.strip(l[0])
            return p
    return None

            
def on_execute_button_pressed():
    execute(use_terminal = False)

def on_execute_in_terminal_button_pressed():
    execute(use_terminal = True)

def execute(use_terminal):
    global last_usage_time_of_program, usage_count_of_program, last_clicked_program
    global arguments_of_program, last_usage_time_of_dir, usage_count_of_dir
    global program_runs_in_terminal, files_whose_light_bulb_is_pressed

    cmd = str(commandEntry.text())
    if len(cmd)>0:
        l = split_spaces_but_not_backslash_space(cmd)
        if len(l)>0:
            p = string.strip(l[0]) # l[0] is the program, and can be a file with path too.
            if ( p in programs
                 or (os.path.isfile(p)
                     and os.access(p, os.X_OK))):
                usage_count_of_program[p] = usage_count_of_program.get(p, 0) +1
                last_usage_time_of_program[p] = datetime.datetime.now()
                # BEGIN increase stats for files
                l2 = [ undo_make_string_bash_compatible(a) for a in l]
                arguments = l2[1:]
                files = [f for f in l2 if os.path.exists(f)]
                for f in files:
                    usage_count_of_dir[f] = usage_count_of_dir.get(f,0) +1
                    last_usage_time_of_dir[f] = datetime.datetime.now()
                # END

                # BEGIN set stats for p
                if arguments_of_program.get(p, None) == None:
                    arguments_of_program[p] = []
                for a in arguments:
                    fi = find(arguments_of_program[p],
                              lambda pair: pair.name == a)
                    if fi ==None:
                        arguments_of_program[p].append( argument_info(a,1, datetime.datetime.now()))
                    else:
                        arguments_of_program[p].remove( fi)
                        arguments_of_program[p].append(
                            argument_info(a, fi.usage_count +1,
                                          datetime.datetime.now()))
                if use_terminal:
                    program_runs_in_terminal[p] = "yes"
                else:
                    program_runs_in_terminal[p] = "no"

                if  types_manageable_by_program.get(p, None) == None:
                    types_manageable_by_program[p] = Set() # the empty set
                for f in files:
                    e = ext(f)
                    if e != "":
                        types_manageable_by_program[p].add(e)
                    else:
                        if os.path.isdir(f):
                            types_manageable_by_program[p].add("<directory>")
                        else:
                            types_manageable_by_program[p].add("")

                            

                # END
                filterQuickText['program'].setText("")
                commandEntry.setText("")
                commandEntry.setFamily("mono") # seems necessary, because it resets itself
                last_clicked_program = None
                print cmd
                if use_terminal:
                    os.system(config.string_to_execute_command_in_terminal + " "+ cmd + " & "
                              #+; echo \"You can close the window now\" # does not work!
                              )
                else:
                    os.system(cmd + " &")
                files_whose_light_bulb_is_pressed = []
                redraw_programs()
                redraw_files()
                redraw_arguments()

            else:
                QMessageBox.critical(mainVBox
                                  , "Program not recognized"
                                  , "The program \"" + p + "\" has been rejected because it was not in the path, \nit was not an executable file and it was not a known built-in command.")


QObject.connect(execute_button, SIGNAL("pressed()"), on_execute_button_pressed)
QObject.connect(execute_in_term_button, SIGNAL("pressed()"), on_execute_in_terminal_button_pressed)

initial_message =  _("To begin a new command, click a program or a file's light-bulb icon.")
help_hbox = QHBox(mainVBox)
help_hbox.setMargin(spacing)
help_hbox.setPaletteBackgroundColor(help_hbox.palette().active().mid())
help_hbox.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)
help_hbox.setSpacing(spacing)
if help_bar_visible:
    help_hbox.show()
else:
    help_hbox.hide()
help_label = FlashingLabel( initial_message, help_hbox)
help_label.setAlignment(Qt.AlignRight |Qt.AlignVCenter)
help_label.startFlashing(config.gFlashingTimeForHelpMessage
                         , resetOriginalColorsWhenFinished = True)
message_explaining_programs_and_narrowing = _("The \"programs\" list below contains the programs currently installed\nin your \"path\". You must always click a program first, to begin a\ncommand; only then you will be able to click files and other\narguments. However, if you want to open a file which you already see,\nand don't want to click a program first, because the program list is\ntoo long, you can click on the \"light bulb\" icon besides the file:\nthis will suggest the programs you can use to view that file, thus \nallowing a document-based interaction.")
QToolTip.add(help_label, message_explaining_programs_and_narrowing)
more_help_button = QPushButton(QIconSet(QPixmap(config.imgDir + "gnome-status.png"))
                               ,_("More details"), help_hbox)
more_help_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(more_help_button, _("Gives you more informations."))
def on_more_details_button_clicked():
    QMessageBox.information(mainVBox, _("More details")

, _("Read the tooltips for more details. Tooltips are messages that appear\nwhen you place the mouse over widgets and wait some time without\nmoving it.\n\nFor example, close this window, then place the mouse over\nthe flashing message. You'll have additional details.")
    )
QObject.connect(more_help_button, SIGNAL("clicked()")
                , on_more_details_button_clicked)

hide_help_button=QPushButton(QIconSet(QPixmap(config.imgDir + "delete.png"))
                             , _("Hide")
                             , help_hbox)
hide_help_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(hide_help_button, _("Hides this help bar to gain more space. You will be able to activate it again \nwith a toolbar button."))
def on_hide_help():
    help_hbox.hide()
    global help_bar_visible
    help_bar_visible = False
    show_help_button.show()
    show_help_button.startFlashing(config.gFlashingTimeForButtons *4
                                   , resetOriginalColorsWhenFinished = True)
QObject.connect(hide_help_button, SIGNAL("clicked()")
                , on_hide_help)
mainGrid = QGrid(3* resolution_multiplier, mainVBox)
mainGrid.setSpacing(spacing)
mainGrid.setMargin(spacing)
mainGrid.show()

hBoxSpacing = 3 * resolution_multiplier

# FIRST ROW
# create three hBox, indexed by panel
for k in ['file','program',  'argument']:
    hBox[k] = QHBox(mainGrid)
    hBox[k].show()
    hBox[k].setSpacing(hBoxSpacing)

    if k != 'file':
        icon[k] = QLabel(hBox[k])
        icon[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        icon[k].show()

set_current_dir_button = QPushButton(QIconSet(QPixmap(config.imgDir + "folder.png"))
                                     ,"",hBox['file'])
set_current_dir_button.show()
set_current_dir_button.setToggleButton(True)
def on_current_dir_changed():
    os.chdir(cur_dir)
QObject.connect(set_current_dir_button, SIGNAL("pressed()"), on_current_dir_changed)
set_current_dir_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
QToolTip.add(set_current_dir_button,

_("Set this directory as the current bash \"working directory\". This is needed\nbecause some programs such as cvs behave in a different way according to\nthe current directory. Using this button is like typing \"cd <dir>\" before \nexecuting a command."))

locationButton = WrappingButton('text will change', hBox['file'], data = None, toggle = False
                                , scalable_values = scalable)
QObject.connect(locationButton, PYSIGNAL("pressed"), onLocationButtonPressed)
locationButton.show()

QToolTip.add(locationButton,

_("This button shows the current directory, and the list below shows\nthe files in the current directory.\nThis button can also be clicked: if you do, the name of the current\ndirectory will be copied to the command area. This button can only be\nclicked if at least a program has been inserted in the command area."))


narrow_dir_button = QPushButton(QIconSet(QPixmap(config.imgDir + "bulb.png")),
                                "",
                                hBox['file'])
narrow_dir_button.setToggleButton(True)
narrow_dir_button.show()
narrow_dir_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
narrow_dir_button.setMinimumWidth(40 * resolution_multiplier)
# def on_narrow_dir_button_released():
#     global files_whose_light_bulb_is_pressed
#     extens_before = extensions_selected_for_narrowing()
#     files_whose_light_bulb_is_pressed.remove(cur_dir)
#     extens_after = extensions_selected_for_narrowing()
#     if len(extens_after) < len(extens_before):
#         redraw_programs()
#         listBox['program'].startFlashing(config.gFlashingTimeForList,
#                                          resetOriginalColorsWhenFinished = True)
def on_narrow_dir_button_toggled(on):
    global files_whose_light_bulb_is_pressed
    if on:
        extens_before = extensions_selected_for_narrowing()
        files_whose_light_bulb_is_pressed.append(cur_dir)
        extens_after = extensions_selected_for_narrowing()
        if len(extens_after) > len(extens_before):
            redraw_programs()
            listBox['program'].startFlashing(config.gFlashingTimeForList,
                                             resetOriginalColorsWhenFinished = True)
    else:
        extens_before = extensions_selected_for_narrowing()
        files_whose_light_bulb_is_pressed.remove(cur_dir)
        extens_after = extensions_selected_for_narrowing()
        if len(extens_after) < len(extens_before):
            redraw_programs()
            listBox['program'].startFlashing(config.gFlashingTimeForList,
                                             resetOriginalColorsWhenFinished = True)
QObject.connect(narrow_dir_button, SIGNAL("toggled(bool)"), on_narrow_dir_button_toggled)

QToolTip.add(narrow_dir_button,

_("Light bulbs like this one can be found beside each file. Clicking a\nlight bulb means \"Suggest me a program to open this file\". At the\nbeginning, OneFinger will have nothing to suggest, but as time goes\nby, OneFinger will learn which programs can open which files, by\nwatching the way you work."))

program_label = FlashingLabel(_('Programs'), hBox['program'])
program_label.show()
program_label.startFlashing(config.gFlashingTimeForHelpMessage
                            , resetOriginalColorsWhenFinished = True)
icon['program'].setPixmap(QPixmap( config.imgDir + "person.png"))




QToolTip.add(program_label, message_explaining_programs_and_narrowing)

argument_label = QLabel(_('Arguments'), hBox['argument'])
argument_label.show()
icon['argument'].setPixmap(QPixmap( config.imgDir + "options.png"))

QToolTip.add(argument_label,
        
_("The list below contains the \"arguments\" you passed to the active\nprogram in your previous commands. Arguments can be files, or options\nsuch as \"--human-readable\". This panel exists mainly because it's\ndifficult to remember some options you use very rarely, like \n\"update-alternatives --config x-session-manager\". This way, you\ndon't have to read the man page each time you want to use a rarely use\ncommand."))


# SECOND ROW
def on_list_item_button_pressed(list, item, button_meaning):
    if button_meaning == "narrow":
        global files_whose_light_bulb_is_pressed
        extens_before = extensions_selected_for_narrowing()
        files_whose_light_bulb_is_pressed.append(item.data)
        extens_after = extensions_selected_for_narrowing()
        if len(extens_after) > len(extens_before):
            redraw_programs()
            listBox['program'].startFlashing(config.gFlashingTimeForList,
                                             resetOriginalColorsWhenFinished = True)

    else:
        raise "by god"
    
def on_list_item_button_released(list, item, button_meaning):
    if button_meaning == "narrow":
        global files_whose_light_bulb_is_pressed
        extens_before = extensions_selected_for_narrowing()
        files_whose_light_bulb_is_pressed.remove(item.data)        
        extens_after = extensions_selected_for_narrowing()
        if len(extens_after) < len(extens_before):
            redraw_programs()
            listBox['program'].startFlashing(config.gFlashingTimeForList,
                                             resetOriginalColorsWhenFinished = True)

    else:
        raise "by god"
    
def on_list_item_button_clicked(list, item, button_meaning):
    k = find([ 'file', 'program', 'argument'], lambda x: listBox[x] is list)
    if k == 'file':
        if button_meaning == "enter directory":
            global recent_dirs, cur_dir
            recent_dirs = add_cur_dir_to_recent_dirs(recent_dirs)
            cur_dir = item.data
            filterQuickText[k].clear()
            usage_count_of_dir[cur_dir] = usage_count_of_dir.get(cur_dir, 0) + 1
            last_usage_time_of_dir[cur_dir] = datetime.datetime.now()
            recent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                            resetOriginalColorsWhenFinished = True)
            rebuild_recent_locations_menu(recent_dirs)
            #don't change; getSortingCriterion is an english string:
            cr = _(sorting_criterion_for_dir.get(cur_dir, "Recent usage")) 
            sortBox[k].setCurrentText(cr)
            redraw_files()
            recent_dirsButton.startFlashing(config.gFlashingTimeForButtons,
                                            resetOriginalColorsWhenFinished = True)
            listBox['file'].startFlashing(config.gFlashingTimeForList,
                                          resetOriginalColorsWhenFinished = True)

        else:
            raise "by god"
    
    elif k == 'program':
        global last_clicked_program
        last_clicked_program = item.data
        redraw_arguments()
        listBox['argument'].startFlashing(config.gFlashingTimeForList,
                                          resetOriginalColorsWhenFinished = True)



def make_string_bash_compatible(f):
    t = string.replace(f, ' ', '\\ ')
    t = string.replace(t, '(', '\\(')
    t = string.replace(t, '[', '\\[')
    t = string.replace(t, ')', '\\)')
    t = string.replace(t, ']', '\\]')
    t = string.replace(t, '\'', '\\\'')
    t = string.replace(t, '\"', '\\\"')
    return t

def undo_make_string_bash_compatible(f):
    t = string.replace(f, '\\ ', ' ')
    t = string.replace(t, '\\(', '(')
    t = string.replace(t, '\\[', '[')
    t = string.replace(t, '\\)', ')')
    t = string.replace(t, '\\]', ']')
    t = string.replace(t, '\\\'', '\'')
    t = string.replace(t, '\\\"', '\"')
    return t
    

def split_spaces_but_not_backslash_space(s):
    ret = []
    cur_word = ""
    advancing_to_next_nonspace=True
    for i in range(0, len(s)):
        if advancing_to_next_nonspace:
            if s[i] == " ":
                pass
            else:
                advancing_to_next_nonspace = False
                cur_word += s[i]
        else:
            if s[i] == " " and s[i-1] != "\\":
                ret.append(cur_word)
                cur_word = ""
                advancing_to_next_nonspace = True
            else:
                cur_word += s[i]

    if len(cur_word)>0:
        ret.append(cur_word)
    return ret

# full = "hello \\(world sss\\ ss"
# print full
# lis = split_spaces_but_not_backslash_space(full)
# print "result"
# for x in lis:
#     print x


    
def on_list_item_clicked(box, item):
    commandEntry.startFlashing(3, True)
    t = make_string_bash_compatible(item.data)
    commandEntry.insert(t + "  ")
    commandEntry.setFocus()


    k = find([ 'file', 'program','argument'], lambda x: listBox[x] is box)
    if k == 'program':
        global last_clicked_program
        last_clicked_program = item.data
        # BEGIN DON'T DO THIS! It is the job of on_text_changed to
        # flash the file listbox; doing this here would only result in
        # duplicated redrawing of the file list.
        #
        #if the_string_specifies_a_program(...) == False:
        #listBox['file'].startFlashing(config.gFlashingTimeForList,
        #                             resetOriginalColorsWhenFinished = True)
        #redraw_files(centering = True)
        #
        #END
        listBox['argument'].startFlashing(config.gFlashingTimeForList,
                                          resetOriginalColorsWhenFinished = True)
        redraw_arguments()



for k in [ "file", "program", "argument"]:
    listBox[k] = MyListBox(frameVisible = False, parent = mainGrid
                           , scalable_values = scalable)
    listBox[k].show()
    QObject.connect(listBox[k], PYSIGNAL("item_button_clicked"), on_list_item_button_clicked)
    QObject.connect(listBox[k], PYSIGNAL("item_button_pressed"), on_list_item_button_pressed)
    QObject.connect(listBox[k], PYSIGNAL("item_button_released"), on_list_item_button_released)
    QObject.connect(listBox[k], PYSIGNAL("item_clicked"), on_list_item_clicked)



# THIRD ROW
for k in [ "file", "program","argument"]:
    filterBox[k] = QHBox(mainGrid)
    filterBox[k].show()
    filterBox[k].setSpacing(3 * resolution_multiplier)

    filterClearButton[k] = QPushButton(filterBox[k])
    filterClearButton[k].setIconSet(QIconSet(QPixmap(config.imgDir + 'delete-right.png')))
    filterClearButton[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    filterQuickText[k] = MyLineEdit(filterBox[k])
    QObject.connect(filterQuickText[k], PYSIGNAL("timer"), onFilterKeyTimer)
    QObject.connect(filterQuickText[k], SIGNAL("returnPressed()"), onEnterPressedOnFilter)

    filterButton[k] = QPushButton(QIconSet(QPixmap(config.imgDir + "mbuto.png")),
                                  _("Recent filters"), filterBox[k])
    filterButton[k].show()

    filterMenu[k] = MyPopupMenu(filterButton[k])
    QObject.connect(filterMenu[k], PYSIGNAL("itemActivated"), onRecentFilterClicked)
    filterButton[k].setPopup(filterMenu[k])

QToolTip.add(filterClearButton['file'],
 _("This button clears the filter area to the right. This means you will see all the \nfiles that were previously hidden."))
QToolTip.add(filterClearButton['program'],
 _("This button clears the filter area to the right. This means you will see all the \nprograms that were previously hidden."))
QToolTip.add(filterClearButton['argument'],
 _("This button clears the filter area to the right. This means you will see all the \narguments that were previously hidden."))

QToolTip.add(filterQuickText['file'],
 _("This is the file filter area.\nIf you type some text here, only those files whose name \ncontains the text will be shown."))
QToolTip.add(filterQuickText['program'],
 _("This is the program filter area.\nIf you type some text here, only those programs whose name \ncontains the text will be shown."))
QToolTip.add(filterQuickText['argument'],
 _("This is the argument filter area.\nIf you type some text here, only those arguments whose name \ncontains the text will be shown."))


QToolTip.add(filterButton['file'],
 _("If you don't use this button often, there's something wrong.\nUse this button to reapply a previous filter to the files list. \nFile filters typically need to be applied again and again: \nthis way you can type them only once and use them forever."))

QToolTip.add(filterButton['program'],
 _("Use this button to reapply a previous filter to the programs list."))

QToolTip.add(filterButton['argument'],
 _("You can use this button to reapply a previous filter to the arguments list."))

rebuildFilterMenus()
QObject.connect(filterClearButton['file'], SIGNAL("clicked()"), clearFileFiltersAndUpdate)
QObject.connect(filterClearButton['program'], SIGNAL("clicked()"), clearProgramFiltersAndUpdate)
QObject.connect(filterClearButton['argument'], SIGNAL("clicked()"), clearArgumentFiltersAndUpdate)

# FOURTH ROW
    
def onSortBoxItemActivated(box, item):
    global sorting_criterion_for_dir, sorting_criterion_for_programs, sorting_criterion_for_arguments
    assert item in ["Recent usage", "Frequent usage", "Date", "Name", "Size"] # sic! don't change. Don't add _().

    k = find(['program', 'file', 'argument'], lambda x: box is sortBox[x])

    if k == 'file' :
        print "setting sorting_criterion_for_dir " + cur_dir + " = " + item
        sorting_criterion_for_dir[cur_dir] = item
        redraw_files()
    if k == 'program':
        sorting_criterion_for_programs = item
        redraw_programs()
    if k== 'argument':
        sorting_criterion_for_arguments = item
        redraw_arguments()
    

for k in ["file", "program", "argument"]:
    sortBox[k] = MyComboBox(mainGrid)
    sortBox[k].setMinimumHeight(35.0 * resolution_multiplier)
    sortBox[k].show()
    sortBox[k].insertItem(config.imgDir + "name.png", text = _("Name"), data = "Name")
    sortBox[k].insertItem(config.imgDir + "stock_timer.png", _("Recent usage"), "Recent usage")
    sortBox[k].insertItem(config.imgDir + "favourites-small.png", _("Frequent usage"), "Frequent usage")
    QObject.connect(sortBox[k], PYSIGNAL("itemActivated"), onSortBoxItemActivated)
sortBox['file'].insertItem(config.imgDir + "size.png", _("Size"), "Size")
sortBox['file'].insertItem(config.imgDir + "clock-small.png", _("Date"), "Date")
sortBox['file'].setCurrentText(_(sorting_criterion_for_dir.get(cur_dir, "Recent usage")))
sortBox['program'].setCurrentText(_(sorting_criterion_for_programs))
sortBox['argument'].setCurrentText(_(sorting_criterion_for_arguments))

QToolTip.add(sortBox['argument'],
 _("You can use this button to change the way arguments are ordered."))
QToolTip.add(sortBox['file'],
             
 _("You can use this button to change the order with which files appear in\nthe above list. Each directory will remember this setting independently. You\nshould almost always leave the default order, \"recent usage\", but\nthere are a few exceptions: for your downloads directory, you should\nset a sorting by date, because this way you can quickly find the\nlatest downloaded file. It can also be useful to sort temporarily by\nsize, if you want to quickly find the largest file in order to burn it\nto CD. The other orderings are useless. In particular, sorting by name\nis useless because it's easier to use the filter box above."))


QToolTip.add(sortBox['program'],

_("Change the order with which programs appear in the above list. There\nis controversy about whether sorting by recent usage is better than\nfrequent usage. I personally prefer recent usage because it still\nkeeps frequent programs above, while taking advantage of \"temporal\ncoherency\". The default is frequent usage, because more people seem\nto like it."))

QToolTip.add(sortBox['argument'],
             _("Change the order with which arguments appear in the above list."))


redraw_files()
redraw_programs()
mainWindow.showMaximized()

rebuild_recent_locations_menu(recent_dirs)
rebuild_important_directories_menu()



applic.exec_loop()

# BEGIN write database
print "Exiting: writing database"
persistence.dumpToFile(targetFile = config.dataDir + 'recent_dirs', var = recent_dirs)
persistence.dumpToFile(targetFile = config.dataDir + 'program_runs_in_terminal', var = program_runs_in_terminal)
persistence.dumpToFile(targetFile = config.dataDir + 'cur_dir', var = cur_dir)
persistence.dumpToFile(targetFile = config.dataDir + 'important_directories', var = important_directories)
persistence.dumpToFile(targetFile = config.dataDir + 'usage_count_of_dir', var = usage_count_of_dir)
persistence.dumpToFile(targetFile = config.dataDir + 'types_manageable_by_program',
                       var = types_manageable_by_program)
persistence.dumpToFile(targetFile = config.dataDir + 'usage_count_of_program', var = usage_count_of_program)
persistence.dumpToFile(targetFile = config.dataDir + 'arguments_of_program', var = arguments_of_program)
persistence.dumpToFile(targetFile = config.dataDir + 'recent_filters', var = recent_filters)
persistence.dumpToFile(targetFile = config.dataDir + 'last_usage_time_of_dir', var = last_usage_time_of_dir)
persistence.dumpToFile(targetFile = config.dataDir + 'last_usage_time_of_program', var = last_usage_time_of_program)
persistence.dumpToFile(targetFile = config.dataDir + 'sorting_criterion_for_dir', var = sorting_criterion_for_dir)
persistence.dumpToFile(targetFile = config.dataDir + 'help_bar_visible', var = help_bar_visible)
persistence.dumpToFile(targetFile = config.dataDir + 'sorting_criterion_for_arguments', var = sorting_criterion_for_arguments)
persistence.dumpToFile(targetFile = config.dataDir + 'sorting_criterion_for_programs', var = sorting_criterion_for_programs)


# END



